@@ -17,6 +17,18 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  */
 
+/**
+ * SECTION:gdict-defbox
+ * @short_description: Display the list of definitions for a word
+ *
+ * The #GdictDefbox widget is a composite widget showing the list of
+ * definitions for a word. It queries the passed #GdictContext and displays
+ * the list of #GdictDefinition<!-- -->s obtained.
+ *
+ * It provides syntax highlighting, clickable links and an embedded find
+ * bar.
+ */
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -28,11 +40,11 @@
 
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
-#include <gtk/gtkbindings.h>
-#include <glib/gi18n.h>
+#include <glib/gi18n-lib.h>
 
 #include "gdict-defbox.h"
 #include "gdict-utils.h"
+#include "gdict-debug.h"
 #include "gdict-private.h"
 #include "gdict-enum-types.h"
 #include "gdict-marshal.h"
@@ -65,8 +77,6 @@ struct _GdictDefboxPrivate
   
   GtkTextBuffer *buffer;
 
-  GtkTooltips *tooltips;
-
   GdictContext *context;
   GSList *definitions;
   
@@ -86,6 +96,10 @@ struct _GdictDefboxPrivate
   guint end_id;
   guint define_id;
   guint error_id;
+  guint hide_timeout;
+
+  GtkTextTag *link_tag;
+  GtkTextTag *visited_link_tag;
 };
 
 enum
@@ -141,53 +155,79 @@ definition_free (Definition *def)
 }
 
 static void
-gdict_defbox_finalize (GObject *object)
+gdict_defbox_dispose (GObject *gobject)
 {
-  GdictDefbox *defbox = GDICT_DEFBOX (object);
+  GdictDefbox *defbox = GDICT_DEFBOX (gobject);
   GdictDefboxPrivate *priv = defbox->priv;
-  
+
   if (priv->start_id)
     {
       g_signal_handler_disconnect (priv->context, priv->start_id);
       g_signal_handler_disconnect (priv->context, priv->end_id);
       g_signal_handler_disconnect (priv->context, priv->define_id);
-    }
-  
-  if (priv->error_id)
-    g_signal_handler_disconnect (priv->context, priv->error_id);
-      
-  if (priv->context)
-    g_object_unref (priv->context);
-  
-  if (priv->database)
-    g_free (priv->database);
-  
-  if (priv->word)
-    g_free (priv->word);
 
-  if (priv->font_name)
-    g_free (priv->font_name);
-  
+      priv->start_id = 0;
+      priv->end_id = 0;
+      priv->define_id = 0;
+    }
+
+  if (priv->error_id)
+    {
+      g_signal_handler_disconnect (priv->context, priv->error_id);
+      priv->error_id = 0;
+    }
+
+  if (priv->context)
+    {
+      g_object_unref (priv->context);
+      priv->context = NULL;
+    }
+
+  if (priv->buffer)
+    {
+      g_object_unref (priv->buffer);
+      priv->buffer = NULL;
+    }
+
+  if (priv->busy_cursor)
+    {
+      gdk_cursor_unref (priv->busy_cursor);
+      priv->busy_cursor = NULL;
+    }
+
+  if (priv->hand_cursor)
+    {
+      gdk_cursor_unref (priv->hand_cursor);
+      priv->hand_cursor = NULL;
+    }
+
+  if (priv->regular_cursor)
+    {
+      gdk_cursor_unref (priv->regular_cursor);
+      priv->regular_cursor = NULL;
+    }
+
+  G_OBJECT_CLASS (gdict_defbox_parent_class)->dispose (gobject);
+}
+
+static void
+gdict_defbox_finalize (GObject *object)
+{
+  GdictDefbox *defbox = GDICT_DEFBOX (object);
+  GdictDefboxPrivate *priv = defbox->priv;
+
+  g_free (priv->database);
+  g_free (priv->word);
+  g_free (priv->font_name);
+
   if (priv->definitions)
     {
       g_slist_foreach (priv->definitions, (GFunc) definition_free, NULL);
       g_slist_free (priv->definitions);
-      
+
       priv->definitions = NULL;
     }
-  
-  g_object_unref (priv->tooltips);
-  g_object_unref (priv->buffer);
-  
-  if (priv->busy_cursor)
-    gdk_cursor_unref (priv->busy_cursor);
 
-  if (priv->hand_cursor)
-    gdk_cursor_unref (priv->hand_cursor);
-
-  if (priv->regular_cursor)
-    gdk_cursor_unref (priv->regular_cursor);
-  
   G_OBJECT_CLASS (gdict_defbox_parent_class)->finalize (object);
 }
 
@@ -204,7 +244,7 @@ set_gdict_context (GdictDefbox  *defbox,
     {
       if (priv->start_id)
         {
-          _gdict_debug ("Removing old context handlers\n");
+          GDICT_NOTE (DEFBOX, "Removing old context handlers");
           
           g_signal_handler_disconnect (priv->context, priv->start_id);
           g_signal_handler_disconnect (priv->context, priv->define_id);
@@ -222,7 +262,7 @@ set_gdict_context (GdictDefbox  *defbox,
           priv->error_id = 0;
         }
 
-      _gdict_debug ("Removing old context\n");
+      GDICT_NOTE (DEFBOX, "Removing old context");
       
       g_object_unref (G_OBJECT (priv->context));
     }
@@ -237,7 +277,7 @@ set_gdict_context (GdictDefbox  *defbox,
       return;
     }
 
-  _gdict_debug ("Setting new context\n");
+  GDICT_NOTE (DEFBOX, "Setting new context");
     
   priv->context = context;
   g_object_ref (G_OBJECT (priv->context));
@@ -254,6 +294,9 @@ gdict_defbox_set_property (GObject      *object,
   
   switch (prop_id)
     {
+    case PROP_WORD:
+      gdict_defbox_lookup (defbox, g_value_get_string (value));
+      break;
     case PROP_CONTEXT:
       set_gdict_context (defbox, g_value_get_object (value));
       break;
@@ -281,6 +324,9 @@ gdict_defbox_get_property (GObject    *object,
   
   switch (prop_id)
     {
+    case PROP_WORD:
+      g_value_set_string (value, priv->word);
+      break;
     case PROP_CONTEXT:
       g_value_set_object (value, priv->context);
       break;
@@ -296,6 +342,715 @@ gdict_defbox_get_property (GObject    *object,
     }
 }
 
+/*
+ * this code has been copied from gtksourceview; it's the implementation
+ * for case-insensitive search in a GtkTextBuffer. this is non-trivial, as
+ * searches on a utf-8 text stream involve a norm(casefold(norm(utf8)))
+ * operation which can be costly on large buffers. luckily for us, it's
+ * not the case on a set of definitions.
+ */
+
+#define GTK_TEXT_UNKNOWN_CHAR 0xFFFC
+
+/* this function acts like g_utf8_offset_to_pointer() except that if it finds a
+ * decomposable character it consumes the decomposition length from the given
+ * offset.  So it's useful when the offset was calculated for the normalized
+ * version of str, but we need a pointer to str itself. */
+static const gchar *
+pointer_from_offset_skipping_decomp (const gchar *str, gint offset)
+{
+	gchar *casefold, *normal;
+	const gchar *p, *q;
+
+	p = str;
+	while (offset > 0)
+	{
+		q = g_utf8_next_char (p);
+		casefold = g_utf8_casefold (p, q - p);
+		normal = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+		offset -= g_utf8_strlen (normal, -1);
+		g_free (casefold);
+		g_free (normal);
+		p = q;
+	}
+	return p;
+}
+
+static gboolean
+exact_prefix_cmp (const gchar *string,
+		  const gchar *prefix,
+		  guint        prefix_len)
+{
+	GUnicodeType type;
+
+	if (strncmp (string, prefix, prefix_len) != 0)
+		return FALSE;
+	if (string[prefix_len] == '\0')
+		return TRUE;
+
+	type = g_unichar_type (g_utf8_get_char (string + prefix_len));
+
+	/* If string contains prefix, check that prefix is not followed
+	 * by a unicode mark symbol, e.g. that trailing 'a' in prefix
+	 * is not part of two-char a-with-hat symbol in string. */
+	return type != G_UNICODE_COMBINING_MARK &&
+		type != G_UNICODE_ENCLOSING_MARK &&
+		type != G_UNICODE_NON_SPACING_MARK;
+}
+
+static const gchar *
+utf8_strcasestr (const gchar *haystack, const gchar *needle)
+{
+	gsize needle_len;
+	gsize haystack_len;
+	const gchar *ret = NULL;
+	gchar *p;
+	gchar *casefold;
+	gchar *caseless_haystack;
+	gint i;
+
+	g_return_val_if_fail (haystack != NULL, NULL);
+	g_return_val_if_fail (needle != NULL, NULL);
+
+	casefold = g_utf8_casefold (haystack, -1);
+	caseless_haystack = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+	g_free (casefold);
+
+	needle_len = g_utf8_strlen (needle, -1);
+	haystack_len = g_utf8_strlen (caseless_haystack, -1);
+
+	if (needle_len == 0)
+	{
+		ret = (gchar *)haystack;
+		goto finally_1;
+	}
+
+	if (haystack_len < needle_len)
+	{
+		ret = NULL;
+		goto finally_1;
+	}
+
+	p = (gchar*)caseless_haystack;
+	needle_len = strlen (needle);
+	i = 0;
+
+	while (*p)
+	{
+		if (exact_prefix_cmp (p, needle, needle_len))
+		{
+			ret = pointer_from_offset_skipping_decomp (haystack, i);
+			goto finally_1;
+		}
+
+		p = g_utf8_next_char (p);
+		i++;
+	}
+
+finally_1:
+	g_free (caseless_haystack);
+
+	return ret;
+}
+
+static const gchar *
+utf8_strrcasestr (const gchar *haystack, const gchar *needle)
+{
+	gsize needle_len;
+	gsize haystack_len;
+	const gchar *ret = NULL;
+	gchar *p;
+	gchar *casefold;
+	gchar *caseless_haystack;
+	gint i;
+
+	g_return_val_if_fail (haystack != NULL, NULL);
+	g_return_val_if_fail (needle != NULL, NULL);
+
+	casefold = g_utf8_casefold (haystack, -1);
+	caseless_haystack = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+	g_free (casefold);
+
+	needle_len = g_utf8_strlen (needle, -1);
+	haystack_len = g_utf8_strlen (caseless_haystack, -1);
+
+	if (needle_len == 0)
+	{
+		ret = (gchar *)haystack;
+		goto finally_1;
+	}
+
+	if (haystack_len < needle_len)
+	{
+		ret = NULL;
+		goto finally_1;
+	}
+
+	i = haystack_len - needle_len;
+	p = g_utf8_offset_to_pointer (caseless_haystack, i);
+	needle_len = strlen (needle);
+
+	while (p >= caseless_haystack)
+	{
+		if (exact_prefix_cmp (p, needle, needle_len))
+		{
+			ret = pointer_from_offset_skipping_decomp (haystack, i);
+			goto finally_1;
+		}
+
+		p = g_utf8_prev_char (p);
+		i--;
+	}
+
+finally_1:
+	g_free (caseless_haystack);
+
+	return ret;
+}
+
+static gboolean
+utf8_caselessnmatch (const char *s1, const char *s2,
+		     gssize n1, gssize n2)
+{
+	gchar *casefold;
+	gchar *normalized_s1;
+	gchar *normalized_s2;
+	gint len_s1;
+	gint len_s2;
+	gboolean ret = FALSE;
+
+	g_return_val_if_fail (s1 != NULL, FALSE);
+	g_return_val_if_fail (s2 != NULL, FALSE);
+	g_return_val_if_fail (n1 > 0, FALSE);
+	g_return_val_if_fail (n2 > 0, FALSE);
+
+	casefold = g_utf8_casefold (s1, n1);
+	normalized_s1 = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+	g_free (casefold);
+
+	casefold = g_utf8_casefold (s2, n2);
+	normalized_s2 = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+	g_free (casefold);
+
+	len_s1 = strlen (normalized_s1);
+	len_s2 = strlen (normalized_s2);
+
+	if (len_s1 < len_s2)
+		goto finally_2;
+
+	ret = (strncmp (normalized_s1, normalized_s2, len_s2) == 0);
+
+finally_2:
+	g_free (normalized_s1);
+	g_free (normalized_s2); 
+
+	return ret;
+}
+
+/* FIXME: total horror */
+static gboolean
+char_is_invisible (const GtkTextIter *iter)
+{
+	GSList *tags;
+	gboolean invisible = FALSE;
+	tags = gtk_text_iter_get_tags (iter);
+	while (tags)
+	{
+		gboolean this_invisible, invisible_set;
+		g_object_get (tags->data, "invisible", &this_invisible, 
+			      "invisible-set", &invisible_set, NULL);
+		if (invisible_set)
+			invisible = this_invisible;
+		tags = g_slist_delete_link (tags, tags);
+	}
+	return invisible;
+}
+
+static void
+forward_chars_with_skipping (GtkTextIter *iter,
+			     gint         count,
+			     gboolean     skip_invisible,
+			     gboolean     skip_nontext,
+			     gboolean     skip_decomp)
+{
+	gint i;
+
+	g_return_if_fail (count >= 0);
+
+	i = count;
+
+	while (i > 0)
+	{
+		gboolean ignored = FALSE;
+
+		/* minimal workaround to avoid the infinite loop of bug #168247.
+		 * It doesn't fix the problemjust the symptom...
+		 */
+		if (gtk_text_iter_is_end (iter))
+			return;
+
+		if (skip_nontext && gtk_text_iter_get_char (iter) == GTK_TEXT_UNKNOWN_CHAR)
+			ignored = TRUE;
+
+		/* FIXME: char_is_invisible() gets list of tags for each char there,
+		   and checks every tag. It doesn't sound like a good idea. */
+		if (!ignored && skip_invisible && char_is_invisible (iter))
+			ignored = TRUE;
+
+		if (!ignored && skip_decomp)
+		{
+			/* being UTF8 correct sucks; this accounts for extra
+			   offsets coming from canonical decompositions of
+			   UTF8 characters (e.g. accented characters) which 
+			   g_utf8_normalize() performs */
+			gchar *normal;
+			gchar buffer[6];
+			gint buffer_len;
+
+			buffer_len = g_unichar_to_utf8 (gtk_text_iter_get_char (iter), buffer);
+			normal = g_utf8_normalize (buffer, buffer_len, G_NORMALIZE_NFD);
+			i -= (g_utf8_strlen (normal, -1) - 1);
+			g_free (normal);
+		}
+
+		gtk_text_iter_forward_char (iter);
+
+		if (!ignored)
+			--i;
+	}
+}
+
+static gboolean
+lines_match (const GtkTextIter *start,
+	     const gchar      **lines,
+	     gboolean           visible_only,
+	     gboolean           slice,
+	     GtkTextIter       *match_start,
+	     GtkTextIter       *match_end)
+{
+	GtkTextIter next;
+	gchar *line_text;
+	const gchar *found;
+	gint offset;
+
+	if (*lines == NULL || **lines == '\0')
+	{
+		if (match_start)
+			*match_start = *start;
+		if (match_end)
+			*match_end = *start;
+		return TRUE;
+	}
+
+	next = *start;
+	gtk_text_iter_forward_line (&next);
+
+	/* No more text in buffer, but *lines is nonempty */
+	if (gtk_text_iter_equal (start, &next))
+		return FALSE;
+
+	if (slice)
+	{
+		if (visible_only)
+			line_text = gtk_text_iter_get_visible_slice (start, &next);
+		else
+			line_text = gtk_text_iter_get_slice (start, &next);
+	}
+	else
+	{
+		if (visible_only)
+			line_text = gtk_text_iter_get_visible_text (start, &next);
+		else
+			line_text = gtk_text_iter_get_text (start, &next);
+	}
+
+	if (match_start) /* if this is the first line we're matching */
+	{
+		found = utf8_strcasestr (line_text, *lines);
+	}
+	else
+	{
+		/* If it's not the first line, we have to match from the
+		 * start of the line.
+		 */
+		if (utf8_caselessnmatch (line_text, *lines, strlen (line_text),
+					 strlen (*lines)))
+			found = line_text;
+		else
+			found = NULL;
+	}
+
+	if (found == NULL)
+	{
+		g_free (line_text);
+		return FALSE;
+	}
+
+	/* Get offset to start of search string */
+	offset = g_utf8_strlen (line_text, found - line_text);
+
+	next = *start;
+
+	/* If match start needs to be returned, set it to the
+	 * start of the search string.
+	 */
+	forward_chars_with_skipping (&next, offset, visible_only, !slice, FALSE);
+	if (match_start)
+	{
+		*match_start = next;
+	}
+
+	/* Go to end of search string */
+	forward_chars_with_skipping (&next, g_utf8_strlen (*lines, -1), visible_only, !slice, TRUE);
+
+	g_free (line_text);
+
+	++lines;
+
+	if (match_end)
+		*match_end = next;
+
+	/* pass NULL for match_start, since we don't need to find the
+	 * start again.
+	 */
+	return lines_match (&next, lines, visible_only, slice, NULL, match_end);
+}
+
+static gboolean
+backward_lines_match (const GtkTextIter *start,
+		      const gchar      **lines,
+		      gboolean           visible_only,
+		      gboolean           slice,
+		      GtkTextIter       *match_start,
+		      GtkTextIter       *match_end)
+{
+	GtkTextIter line, next;
+	gchar *line_text;
+	const gchar *found;
+	gint offset;
+
+	if (*lines == NULL || **lines == '\0')
+	{
+		if (match_start)
+			*match_start = *start;
+		if (match_end)
+			*match_end = *start;
+		return TRUE;
+	}
+
+	line = next = *start;
+	if (gtk_text_iter_get_line_offset (&next) == 0)
+	{
+		if (!gtk_text_iter_backward_line (&next))
+			return FALSE;
+	}
+	else
+		gtk_text_iter_set_line_offset (&next, 0);
+
+	if (slice)
+	{
+		if (visible_only)
+			line_text = gtk_text_iter_get_visible_slice (&next, &line);
+		else
+			line_text = gtk_text_iter_get_slice (&next, &line);
+	}
+	else
+	{
+		if (visible_only)
+			line_text = gtk_text_iter_get_visible_text (&next, &line);
+		else
+			line_text = gtk_text_iter_get_text (&next, &line);
+	}
+
+	if (match_start) /* if this is the first line we're matching */
+	{
+		found = utf8_strrcasestr (line_text, *lines);
+	}
+	else
+	{
+		/* If it's not the first line, we have to match from the
+		 * start of the line.
+		 */
+		if (utf8_caselessnmatch (line_text, *lines, strlen (line_text),
+					 strlen (*lines)))
+			found = line_text;
+		else
+			found = NULL;
+	}
+
+	if (found == NULL)
+	{
+		g_free (line_text);
+		return FALSE;
+	}
+
+	/* Get offset to start of search string */
+	offset = g_utf8_strlen (line_text, found - line_text);
+
+	forward_chars_with_skipping (&next, offset, visible_only, !slice, FALSE);
+
+	/* If match start needs to be returned, set it to the
+	 * start of the search string.
+	 */
+	if (match_start)
+	{
+		*match_start = next;
+	}
+
+	/* Go to end of search string */
+	forward_chars_with_skipping (&next, g_utf8_strlen (*lines, -1), visible_only, !slice, TRUE);
+
+	g_free (line_text);
+
+	++lines;
+
+	if (match_end)
+		*match_end = next;
+
+	/* try to match the rest of the lines forward, passing NULL
+	 * for match_start so lines_match will try to match the entire
+	 * line */
+	return lines_match (&next, lines, visible_only,
+			    slice, NULL, match_end);
+}
+
+/* strsplit () that retains the delimiter as part of the string. */
+static gchar **
+breakup_string (const char *string,
+		const char *delimiter,
+		gint        max_tokens)
+{
+	GSList *string_list = NULL, *slist;
+	gchar **str_array, *s, *casefold, *new_string;
+	guint i, n = 1;
+
+	g_return_val_if_fail (string != NULL, NULL);
+	g_return_val_if_fail (delimiter != NULL, NULL);
+
+	if (max_tokens < 1)
+		max_tokens = G_MAXINT;
+
+	s = strstr (string, delimiter);
+	if (s)
+	{
+		guint delimiter_len = strlen (delimiter);
+
+		do
+		{
+			guint len;
+
+			len = s - string + delimiter_len;
+			new_string = g_new (gchar, len + 1);
+			strncpy (new_string, string, len);
+			new_string[len] = 0;
+			casefold = g_utf8_casefold (new_string, -1);
+			g_free (new_string);
+			new_string = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+			g_free (casefold);
+			string_list = g_slist_prepend (string_list, new_string);
+			n++;
+			string = s + delimiter_len;
+			s = strstr (string, delimiter);
+		} while (--max_tokens && s);
+	}
+
+	if (*string)
+	{
+		n++;
+		casefold = g_utf8_casefold (string, -1);
+		new_string = g_utf8_normalize (casefold, -1, G_NORMALIZE_NFD);
+		g_free (casefold);
+		string_list = g_slist_prepend (string_list, new_string);
+	}
+
+	str_array = g_new (gchar*, n);
+
+	i = n - 1;
+
+	str_array[i--] = NULL;
+	for (slist = string_list; slist; slist = slist->next)
+		str_array[i--] = slist->data;
+
+	g_slist_free (string_list);
+
+	return str_array;
+}
+
+static gboolean
+gdict_defbox_iter_forward_search (const GtkTextIter   *iter,
+                                  const gchar         *str,
+                                  GtkTextIter         *match_start,
+                                  GtkTextIter         *match_end,
+                                  const GtkTextIter   *limit)
+{
+  gchar **lines = NULL;
+  GtkTextIter match;
+  gboolean retval = FALSE;
+  GtkTextIter search;
+
+  g_return_val_if_fail (iter != NULL, FALSE);
+  g_return_val_if_fail (str != NULL, FALSE);
+
+  if (limit && gtk_text_iter_compare (iter, limit) >= 0)
+    return FALSE;
+
+  if (*str == '\0')
+    {
+      /* If we can move one char, return the empty string there */
+      match = *iter;
+
+      if (gtk_text_iter_forward_char (&match))
+        {
+          if (limit && gtk_text_iter_equal (&match, limit))
+            return FALSE;
+
+          if (match_start)
+            *match_start = match;
+
+          if (match_end)
+            *match_end = match;
+
+          return TRUE;
+        }
+      else
+        return FALSE;
+    }
+
+  /* locate all lines */
+  lines = breakup_string (str, "\n", -1);
+
+  search = *iter;
+
+  /* This loop has an inefficient worst-case, where
+   * gtk_text_iter_get_text () is called repeatedly on
+   * a single line.
+   */
+  do
+    {
+      GtkTextIter end;
+      gboolean res;
+
+      if (limit && gtk_text_iter_compare (&search, limit) >= 0)
+        break;
+
+      res = lines_match (&search, (const gchar**)lines,
+                         TRUE, FALSE,
+                         &match, &end);
+      if (res)
+        {
+          if (limit == NULL ||
+              (limit && gtk_text_iter_compare (&end, limit) <= 0))
+            {
+              retval = TRUE;
+              
+              if (match_start)
+                *match_start = match;
+
+              if (match_end)
+                *match_end = end;
+            }
+
+          break;
+        }
+    } while (gtk_text_iter_forward_line (&search));
+
+  g_strfreev ((gchar**) lines);
+
+  return retval;
+}
+
+static gboolean
+gdict_defbox_iter_backward_search (const GtkTextIter   *iter,
+                                   const gchar         *str,
+                                   GtkTextIter         *match_start,
+                                   GtkTextIter         *match_end,
+                                   const GtkTextIter   *limit)
+{
+  gchar **lines = NULL;
+  GtkTextIter match;
+  gboolean retval = FALSE;
+  GtkTextIter search;
+
+  g_return_val_if_fail (iter != NULL, FALSE);
+  g_return_val_if_fail (str != NULL, FALSE);
+
+  if (limit && gtk_text_iter_compare (iter, limit) <= 0)
+    return FALSE;
+
+  if (*str == '\0')
+    {
+      /* If we can move one char, return the empty string there */
+      match = *iter;
+
+      if (gtk_text_iter_backward_char (&match))
+        {
+          if (limit && gtk_text_iter_equal (&match, limit))
+            return FALSE;
+
+          if (match_start)
+            *match_start = match;
+
+          if (match_end)
+            *match_end = match;
+
+          return TRUE;
+        }
+      else
+        return FALSE;
+    }
+
+  /* locate all lines */
+  lines = breakup_string (str, "\n", -1);
+
+  search = *iter;
+
+  /* This loop has an inefficient worst-case, where
+   * gtk_text_iter_get_text () is called repeatedly on
+   * a single line.
+   */
+  while (TRUE)
+    {
+      GtkTextIter end;
+      gboolean res;
+
+      if (limit && gtk_text_iter_compare (&search, limit) <= 0)
+        break;
+
+      res = backward_lines_match (&search, (const gchar**)lines,
+                                  TRUE, FALSE,
+                                  &match, &end);
+      if (res)
+        {
+           if (limit == NULL ||
+               (limit && gtk_text_iter_compare (&end, limit) > 0))
+             {
+               retval = TRUE;
+               
+               if (match_start)
+                 *match_start = match;
+
+               if (match_end)
+                 *match_end = end;
+
+             }
+           
+           break;
+        }
+
+      if (gtk_text_iter_get_line_offset (&search) == 0)
+        {
+          if (!gtk_text_iter_backward_line (&search))
+            break;
+        }
+      else
+        gtk_text_iter_set_line_offset (&search, 0);
+    }
+
+  g_strfreev ((gchar**) lines);
+
+  return retval;
+}
+
 static gboolean
 gdict_defbox_find_backward (GdictDefbox *defbox,
 			    const gchar *text)
@@ -318,11 +1073,9 @@ gdict_defbox_find_backward (GdictDefbox *defbox,
   else
     iter = end_iter;
   
-  res = gtk_text_iter_backward_search (&iter, text,
-  				       GTK_TEXT_SEARCH_VISIBLE_ONLY | GTK_TEXT_SEARCH_TEXT_ONLY,
-  				       &match_start,
-  				       &match_end,
-  				       NULL);
+  res = gdict_defbox_iter_backward_search (&iter, text,
+                                           &match_start, &match_end,
+                                           NULL);
   if (res)
     {
       gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (priv->text_view),
@@ -343,6 +1096,21 @@ gdict_defbox_find_backward (GdictDefbox *defbox,
   return FALSE;
 }
 
+static gboolean
+hide_find_pane (gpointer user_data)
+{
+  GdictDefbox *defbox = user_data;
+
+  gtk_widget_hide (defbox->priv->find_pane);
+  defbox->priv->show_find = FALSE;
+  
+  gtk_widget_grab_focus (defbox->priv->text_view);
+  
+  defbox->priv->hide_timeout = 0;
+
+  return FALSE;
+}
+
 static void
 find_prev_clicked_cb (GtkWidget *widget,
 		      gpointer   user_data)
@@ -369,6 +1137,12 @@ find_prev_clicked_cb (GtkWidget *widget,
       
       g_free (str);
     }
+
+  if (priv->hide_timeout)
+    {
+      g_source_remove (priv->hide_timeout);
+      priv->hide_timeout = g_timeout_add (5000, hide_find_pane, defbox);
+    }
 }
 
 static gboolean
@@ -407,11 +1181,10 @@ gdict_defbox_find_forward (GdictDefbox *defbox,
 	iter = start_iter;
     }
   
-  res = gtk_text_iter_forward_search (&iter, text,
-  				      GTK_TEXT_SEARCH_VISIBLE_ONLY | GTK_TEXT_SEARCH_TEXT_ONLY,
-  				      &match_start,
-  				      &match_end,
-  				      NULL);
+  res = gdict_defbox_iter_forward_search (&iter, text,
+                                          &match_start,
+                                          &match_end,
+                                          NULL);
   if (res)
     {
       gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (priv->text_view),
@@ -458,6 +1231,12 @@ find_next_clicked_cb (GtkWidget *widget,
       
       g_free (str);
     }
+
+  if (priv->hide_timeout)
+    {
+      g_source_remove (priv->hide_timeout);
+      priv->hide_timeout = g_timeout_add (5000, hide_find_pane, defbox);
+    }
 }
 
 static void
@@ -488,6 +1267,24 @@ find_entry_changed_cb (GtkWidget *widget,
     }
 
   g_free (text);
+
+  if (priv->hide_timeout)
+    {
+      g_source_remove (priv->hide_timeout);
+      priv->hide_timeout = g_timeout_add (5000, hide_find_pane, defbox);
+    }
+}
+
+static void
+close_button_clicked (GtkButton *button,
+                      gpointer   data)
+{
+  GdictDefboxPrivate *priv = GDICT_DEFBOX (data)->priv;
+
+  if (priv->hide_timeout)
+    g_source_remove (priv->hide_timeout);
+
+  (void) hide_find_pane (data);
 }
 
 static GtkWidget *
@@ -498,6 +1295,7 @@ create_find_pane (GdictDefbox *defbox)
   GtkWidget *label;
   GtkWidget *sep;
   GtkWidget *hbox1, *hbox2;
+  GtkWidget *button;
  
   priv = defbox->priv;
   
@@ -508,6 +1306,16 @@ create_find_pane (GdictDefbox *defbox)
   gtk_box_pack_start (GTK_BOX (find_pane), hbox1, TRUE, TRUE, 0);
   gtk_widget_show (hbox1);
 
+  button = gtk_button_new ();
+  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
+  gtk_button_set_image (GTK_BUTTON (button),
+                        gtk_image_new_from_stock (GTK_STOCK_CLOSE,
+                                                  GTK_ICON_SIZE_BUTTON));
+  g_signal_connect (button, "clicked",
+                    G_CALLBACK (close_button_clicked), defbox);
+  gtk_box_pack_start (GTK_BOX (hbox1), button, FALSE, FALSE, 0);
+  gtk_widget_show (button);
+
   hbox2 = gtk_hbox_new (FALSE, 12);
   gtk_box_pack_start (GTK_BOX (hbox1), hbox2, TRUE, TRUE, 0);
   gtk_widget_show (hbox2);
@@ -585,14 +1393,16 @@ gdict_defbox_init_tags (GdictDefbox *defbox)
   if (!visited_link_color)
     visited_link_color = &default_visited_link_color;
 
-  gtk_text_buffer_create_tag (priv->buffer, "link",
-                              "underline", PANGO_UNDERLINE_SINGLE,
-                              "foreground-gdk", &link_color,
-                              NULL);
-  gtk_text_buffer_create_tag (priv->buffer, "visited-link",
-                              "underline", PANGO_UNDERLINE_SINGLE,
-                              "foreground-gdk", &visited_link_color,
-                              NULL);
+  priv->link_tag =
+    gtk_text_buffer_create_tag (priv->buffer, "link",
+                                "underline", PANGO_UNDERLINE_SINGLE,
+                                "foreground-gdk", link_color,
+                                NULL);
+  priv->visited_link_tag =
+    gtk_text_buffer_create_tag (priv->buffer, "visited-link",
+                                "underline", PANGO_UNDERLINE_SINGLE,
+                                "foreground-gdk", visited_link_color,
+                                NULL);
 
   if (link_color != &default_link_color)
     gdk_color_free (link_color);
@@ -867,14 +1677,36 @@ static void
 gdict_defbox_style_set (GtkWidget *widget,
 			GtkStyle  *old_style)
 {
-  GdictDefbox *defbox;
-  
-  defbox = GDICT_DEFBOX (widget);
+  GdictDefboxPrivate *priv = GDICT_DEFBOX (widget)->priv;
+  GdkColor *link_color, *visited_link_color;
   
   if (GTK_WIDGET_CLASS (gdict_defbox_parent_class)->style_set)
     GTK_WIDGET_CLASS (gdict_defbox_parent_class)->style_set (widget, old_style);
-  
-  /* TODO handle our style here */
+
+  link_color = visited_link_color = NULL;
+  gtk_widget_style_get (widget,
+                        "link-color", &link_color,
+                        "visited-link-color", &visited_link_color,
+                        NULL);
+  if (!link_color)
+    link_color = &default_link_color;
+
+  if (!visited_link_color)
+    visited_link_color = &default_visited_link_color;
+
+  g_object_set (G_OBJECT (priv->link_tag),
+                "foreground-gdk", link_color,
+                NULL);
+
+  g_object_set (G_OBJECT (priv->visited_link_tag),
+                "foreground-gdk", visited_link_color,
+                NULL);
+
+  if (link_color != &default_link_color)
+    gdk_color_free (link_color);
+
+  if (visited_link_color != &default_visited_link_color)
+    gdk_color_free (visited_link_color);
 }
 
 /* we override the GtkWidget::show_all method since we have widgets
@@ -899,6 +1731,8 @@ gdict_defbox_real_show_find (GdictDefbox *defbox)
   defbox->priv->show_find = TRUE;
   
   gtk_widget_grab_focus (defbox->priv->find_entry);
+
+  defbox->priv->hide_timeout = g_timeout_add (5000, hide_find_pane, defbox);
 }
 
 static void
@@ -922,6 +1756,12 @@ gdict_defbox_real_hide_find (GdictDefbox *defbox)
   defbox->priv->show_find = FALSE;
   
   gtk_widget_grab_focus (defbox->priv->text_view);
+
+  if (defbox->priv->hide_timeout)
+    {
+      g_source_remove (defbox->priv->hide_timeout);
+      defbox->priv->hide_timeout = 0;
+    }
 }
 
 static void
@@ -934,13 +1774,28 @@ gdict_defbox_class_init (GdictDefboxClass *klass)
   gobject_class->constructor = gdict_defbox_constructor;
   gobject_class->set_property = gdict_defbox_set_property;
   gobject_class->get_property = gdict_defbox_get_property;
+  gobject_class->dispose = gdict_defbox_dispose;
   gobject_class->finalize = gdict_defbox_finalize;
   
   widget_class->show_all = gdict_defbox_show_all;
   widget_class->style_set = gdict_defbox_style_set;
-  
+
   /**
-   * GdictDefbox:context
+   * GdictDefbox:word:
+   *
+   * The word to look up.
+   *
+   * Since: 0.10
+   */
+  g_object_class_install_property (gobject_class,
+                                   PROP_WORD,
+                                   g_param_spec_string ("word",
+                                                        "Word",
+                                                        "The word to look up",
+                                                        NULL,
+                                                        G_PARAM_READWRITE));
+  /**
+   * GdictDefbox:context:
    *
    * The #GdictContext object used to get the word definition.
    *
@@ -1076,9 +1931,8 @@ gdict_defbox_init (GdictDefbox *defbox)
   priv->show_find = FALSE;
   priv->is_searching = FALSE;
   priv->is_hovering = FALSE;
-  
-  priv->tooltips = gtk_tooltips_new ();
-  g_object_ref_sink (G_OBJECT (priv->tooltips));
+
+  priv->hide_timeout = 0;
 }
 
 /**
@@ -1115,9 +1969,7 @@ gdict_defbox_new_with_context (GdictContext *context)
 {
   g_return_val_if_fail (GDICT_IS_CONTEXT (context), NULL);
   
-  return g_object_new (GDICT_TYPE_DEFBOX,
-                       "context", context,
-                       NULL);
+  return g_object_new (GDICT_TYPE_DEFBOX, "context", context, NULL);
 }
 
 /**
@@ -1153,15 +2005,9 @@ gdict_defbox_set_context (GdictDefbox  *defbox,
 GdictContext *
 gdict_defbox_get_context (GdictDefbox *defbox)
 {
-  GdictContext *context;
-  
   g_return_val_if_fail (GDICT_IS_DEFBOX (defbox), NULL);
-  
-  g_object_get (defbox, "context", &context, NULL);
-  if (context)
-    g_object_unref (context);
-  
-  return context;
+
+  return defbox->priv->context;
 }
 
 /**
@@ -1169,8 +2015,8 @@ gdict_defbox_get_context (GdictDefbox *defbox)
  * @defbox: a #GdictDefbox
  * @database: a database
  *
- * Sets @database as the database used by the #GdictContext bound to @defbox to
- * query for word definitions.
+ * Sets @database as the database used by the #GdictContext bound to @defbox
+ * to query for word definitions.
  *
  * Since: 0.1
  */
@@ -1178,9 +2024,16 @@ void
 gdict_defbox_set_database (GdictDefbox *defbox,
 			   const gchar *database)
 {
+  GdictDefboxPrivate *priv;
+
   g_return_if_fail (GDICT_IS_DEFBOX (defbox));
 
-  g_object_set (G_OBJECT (defbox), "database", database, NULL);
+  priv = defbox->priv;
+
+  g_free (priv->database);
+  priv->database = g_strdup (database);
+
+  g_object_notify (G_OBJECT (defbox), "database");
 }
 
 /**
@@ -1189,21 +2042,37 @@ gdict_defbox_set_database (GdictDefbox *defbox,
  *
  * Gets the database used by @defbox.  See gdict_defbox_set_database().
  *
- * Return value: the name of a database.  The string is owned by the
- * #GdictDefbox and should not be modified or freed.
+ * Return value: the name of a database. The return string is owned by
+ *   the #GdictDefbox widget and should not be modified or freed.
  *
  * Since: 0.1
  */
 G_CONST_RETURN gchar *
 gdict_defbox_get_database (GdictDefbox *defbox)
 {
-  gchar *database;
-  
   g_return_val_if_fail (GDICT_IS_DEFBOX (defbox), NULL);
 
-  g_object_get (G_OBJECT (defbox), "database", &database, NULL);
+  return defbox->priv->database;
+}
 
-  return database;
+/**
+ * gdict_defbox_get_word:
+ * @defbox: a #GdictDefbox
+ *
+ * Retrieves the word being looked up.
+ *
+ * Return value: the word looked up, or %NULL. The returned string is
+ *   owned by the #GdictDefbox widget and should never be modified or
+ *   freed.
+ *
+ * Since: 0.12
+ */
+G_CONST_RETURN gchar *
+gdict_defbox_get_word (GdictDefbox *defbox)
+{
+  g_return_val_if_fail (GDICT_IS_DEFBOX (defbox), NULL);
+
+  return defbox->priv->word;
 }
 
 /**
@@ -1233,9 +2102,20 @@ gdict_defbox_set_show_find (GdictDefbox *defbox,
     {
       gtk_widget_show_all (priv->find_pane);
       gtk_widget_grab_focus (priv->find_entry);
+
+      if (!priv->hide_timeout)
+        priv->hide_timeout = g_timeout_add (5000, hide_find_pane, defbox);
     }
   else
-    gtk_widget_hide (priv->find_pane);
+    {
+      gtk_widget_hide (priv->find_pane);
+
+      if (priv->hide_timeout)
+        {
+          g_source_remove (priv->hide_timeout);
+          priv->hide_timeout = 0;
+        }
+    }
 }
 
 /**
@@ -1294,9 +2174,6 @@ lookup_end_cb (GdictContext *context,
   				     GTK_TEXT_WINDOW_WIDGET);
   
   gdk_window_set_cursor (window, NULL);
-  
-  g_free (priv->word);
-  priv->word = NULL;
 
   priv->is_searching = FALSE;
 }
@@ -1449,7 +2326,7 @@ gdict_defbox_insert_body (GdictDefbox *defbox,
               buf = g_string_new (NULL);
               next = words[cur++];
 
-              while ((end = g_utf8_strrchr (next, -1, '}')) == NULL)
+              while (next && (end = g_utf8_strrchr (next, -1, '}')) == NULL)
                 {
                   buf = g_string_append (buf, next);
                   buf = g_string_append_c (buf, ' ');
@@ -1461,7 +2338,7 @@ gdict_defbox_insert_body (GdictDefbox *defbox,
 
               next = g_string_free (buf, FALSE);
 
-              if (*end == '}')
+              if (end && *end == '}')
                 {
                   const gchar *rest;
                   gchar *link;
@@ -1708,6 +2585,7 @@ gdict_defbox_lookup (GdictDefbox *defbox,
   				       defbox);
   
   priv->word = g_strdup (word);
+  g_object_notify (G_OBJECT (defbox), "word");
   
   define_error = NULL;
   gdict_context_define_word (priv->context,
@@ -2007,7 +2885,12 @@ gdict_defbox_get_font_name (GdictDefbox *defbox)
  * gdict_defbox_get_selected_word:
  * @defbox: a #GdictDefbox
  *
- * Since: 0.11
+ * Retrieves the selected word from the defbox widget
+ *
+ * Return value: a newly allocated string containing the selected
+ *   word. Use g_free() when done using it.
+ *
+ * Since: 0.12
  */
 gchar *
 gdict_defbox_get_selected_word (GdictDefbox *defbox)
